import jax
import jax.numpy as jnp
import jax.lax as lax

import numpy as np

from basicsr.utils.color_util import bgr2ycbcr
from basicsr.metrics.metric_util import to_y_channel
from iqa.utils.convert_img import rgb2y, preprocess

from absl.testing import absltest, parameterized

from functools import partial
import itertools


jax.config.parse_flags_with_absl()


search_space = {
    'is_single': [True, False],
    'use_cpu': [True, False],
}
search_space_list = list(itertools.product(*search_space.values()))
search_space = [dict(zip(search_space.keys(), v)) for v in search_space_list]


class PreprocessingTest(parameterized.TestCase):
    @parameterized.parameters(*search_space)
    def test_preprocessing(self, is_single, use_cpu):
        if is_single:
            inputs = np.random.randint(0., 256., size=(1024, 1024, 3))
        else:
            inputs = np.random.randint(0., 256., size=(32, 1024, 1024, 3))

        inputs_jax = jnp.array(inputs, dtype=jnp.uint8)
        inputs_bsr = inputs.astype(np.uint8)

        if is_single:  # BasicSR uses BGR2YCbCr to get Y channel. So I reversed the channel.
            y_bsr = to_y_channel(inputs_bsr[..., ::-1])
        else:
            y_bsr = []
            for i in range(inputs_bsr.shape[0]):
                y_bsr.append(to_y_channel(inputs_bsr[i][..., ::-1]))
            y_bsr = np.stack(y_bsr)

        device = jax.devices('cpu' if use_cpu else 'gpu')[0]
        inputs_jax = jax.device_put(inputs_jax, device=device)
        func = jax.jit(partial(preprocess, crop_border=0, to_y=True))
        y_iqa = func(inputs_jax)

        np.testing.assert_allclose(y_bsr.squeeze(), y_iqa.squeeze(), atol=1e-4, rtol=1e-4)


if __name__ == '__main__':
    absltest.main()



import warnings

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from scipy import linalg


def fid(feats_1: np.array, feats_2: np.array, eps: float = 1e-6) -> np.array:
    """
    Current Jax's sqrtm implementation is different from scipy's sqrtm.
    So this implementation is just copy of BasicSR's FID calculation.
    It will be fixed in the future.
    Not directly use BasicSR's FID calculation to prevent issues.
    https://github.com/XPixelGroup/BasicSR/blob/master/basicsr/metrics/fid.py

    Not incorporate feature extraction in this function because original Numpy does not support Jit-compile
    while Flax's module does.
    """
    warnings.warn("Current fid does not support Jit-compile. It will be fixed in the future.")

    mu1 = np.mean(feats_1, axis=0)
    mu2 = np.mean(feats_2, axis=0)

    sigma1 = np.cov(feats_1, rowvar=False)
    sigma2 = np.cov(feats_2, rowvar=False)

    assert mu1.shape == mu2.shape, 'Two mean vectors have different lengths'
    assert sigma1.shape == sigma2.shape, ('Two covariances have different dimensions')

    cov_sqrt, _ = linalg.sqrtm(sigma1 @ sigma2, disp=False)

    # Product might be almost singular
    if not np.isfinite(cov_sqrt).all():
        print('Product of cov matrices is singular. Adding {eps} to diagonal of cov estimates')
        offset = np.eye(sigma1.shape[0]) * eps
        cov_sqrt = linalg.sqrtm((sigma1 + offset) @ (sigma2 + offset))

    # Numerical error might give slight imaginary component
    if np.iscomplexobj(cov_sqrt):
        if not np.allclose(np.diagonal(cov_sqrt).imag, 0, atol=1e-3):
            m = np.max(np.abs(cov_sqrt.imag))
            raise ValueError(f'Imaginary component {m}')
        cov_sqrt = cov_sqrt.real

    mean_diff = mu1 - mu2
    mean_norm = mean_diff @ mean_diff
    trace = np.trace(sigma1) + np.trace(sigma2) - 2 * np.trace(cov_sqrt)
    fid_score = mean_norm + trace

    return fid_score


def fid_jax(
        imgs_1: jnp.ndarray, imgs_2: jnp.ndarray, feats_extractor: nn.Module, eps: float = 1e-6
) -> jnp.ndarray:
    """
    This implementation results different from fid() because of Jax's sqrtm implementation.
    Use this only when Jit-compile must be used.
    """
    warnings.warn("fid_jax function results different from original FID calculation."
                  "Use fid() unless You must use jit-compile.")

    feats_1 = feats_extractor(imgs_1)
    feats_2 = feats_extractor(imgs_2)

    mu1 = jnp.mean(feats_1, axis=0)
    mu2 = jnp.mean(feats_2, axis=0)

    sigma1 = jnp.cov(feats_1, rowvar=False)
    sigma2 = jnp.cov(feats_2, rowvar=False)

    assert mu1.shape == mu2.shape, 'Two mean vectors have different lengths'
    assert sigma1.shape == sigma2.shape, 'Two covariances have different dimensions'

    cov_sqrt = jax.scipy.linalg.sqrtm(sigma1 @ sigma2)

    # Product might be almost singular
    if jnp.isfinite(cov_sqrt).any():
        print('Product of cov matrices is singular. Adding {eps} to diagonal of cov estimates')
        offset = jnp.eye(sigma1.shape[0]) * eps
        cov_sqrt = jax.scipy.linalg.sqrtm((sigma1 + offset) @ (sigma2 + offset))

    # Numerical error might give slight imaginary component
    if jnp.iscomplexobj(cov_sqrt):
        if not jnp.allclose(jnp.diagonal(cov_sqrt).imag, 0, atol=1e-3):
            m = jnp.max(jnp.abs(cov_sqrt.imag))
            raise ValueError(f'Imaginary component {m}')
        cov_sqrt = cov_sqrt.real

    mean_diff = mu1 - mu2
    mean_norm = mean_diff @ mean_diff
    trace = jnp.trace(sigma1) + jnp.trace(sigma2) - 2 * jnp.trace(cov_sqrt)
    fid_score = mean_norm + trace

    return fid_score



import jax.numpy as jnp


def rgb2y(img: jnp.ndarray) -> jnp.ndarray:
    """
    Convert RGB image to Y channel.

    Args:
        img(jnp.ndarray[int, float]): 0 ~ 255 RGB image

    Returns:
        jnp.ndarray[float32]: Y channel image

    """
    img = (img.astype(jnp.float64) / jnp.array(255., dtype=jnp.float64)).astype(jnp.float32)

    out_img = jnp.dot(img, jnp.array([65.481, 128.553, 24.966], dtype=jnp.float64)) \
              + jnp.array(16.0, dtype=jnp.float64)

    return out_img[..., jnp.newaxis].astype(jnp.float32)


def preprocess(img: jnp.ndarray, crop_border: int, to_y: bool) -> jnp.ndarray:
    """
    Preprocessing images for calculate metrics.

    Args:
        img(jnp.ndarray[int, float]): 0 ~ 255 RGB image
        crop_border(int): Crop border size.
        to_y(bool): Whether to only return Y channel

    Returns:
        jnp.ndarray[float64]: Preprocessed image

    """
    if img.ndim == 3:
        img = img[jnp.newaxis, ...]

    if to_y:
        img = rgb2y(img)

    img = img.astype(jnp.float64)

    if crop_border > 0:
        img = img[:, crop_border:-crop_border, crop_border:-crop_border, :]

    return img



import jax
import warnings
from . import metrics
from . import models
from . import utils


jax.config.update('jax_enable_x64', True)
warnings.warn(
    "This library enables jax_enable_x64 by default for precision. "
    "Be aware that other other jax libraries may not be compatible with this."
)



import jax.numpy as jnp

from iqa.utils.convert_img import preprocess


def psnr(
        img1: jnp.ndarray, img2: jnp.ndarray, crop_border: int, test_y: bool
) -> jnp.ndarray:
    """
    Calculate PSNR between two images.

    Args:
        img1: 0 ~ 255 RGB image
        img2: 0 ~ 255 RGB image
        crop_border: Crop border size.
        test_y: Whether to use Y channel for PSNR calculation.

    Returns:
        PSNR value.
    """
    img1 = preprocess(img1, crop_border, test_y)
    img2 = preprocess(img2, crop_border, test_y)

    mse = jnp.mean((img1 - img2) ** 2, axis=(1, 2, 3))
    mask = mse == 0
    return jnp.where(mask, jnp.inf, 10.0 * jnp.log10(255 ** 2 / mse))


class PSNR:
    def __init__(self, crop_border: int, test_y: bool):
        """
        Class for PSNR calculation.
        Args:
            crop_border: Crop border size.
            test_y: Whether to use Y channel for PSNR calculation.
        """
        self.crop_border = crop_border
        self.test_y = test_y

    def __call__(self, img1: jnp.ndarray, img2: jnp.ndarray, **kwargs) -> jnp.ndarray:
        return psnr(img1, img2, self.crop_border, self.test_y)



import setuptools


install_requires = [
    "numpy>=1.12",
    "jax>=0.4.2",
    "tensorflow>=2.0.0",
    "flax>=0.6.0"
]
testing_requires = [
    "basicsr>=1.4.2"
]

setuptools.setup(
    name="iqa-jax",
    version="0.0.1",
    author="dslisleedh",
    author_email="dslisleedh@gmail.com",
    description="IQA library for Jax",
    long_description=open('README.md', 'rt').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/dslisleedh/IQA-jax",
    project_urls={
        "Bug Tracker": "https://github.com/dslisleedh/IQA-jax/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Image Processing"
    ],
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    extras_require={
        "testing": testing_requires
    },
    python_requires=">=3.6",
)



import jax
import jax.numpy as jnp
import jax.lax as lax
import numpy as np

from iqa.utils.convert_img import preprocess


def _get_2d_gaussian_kernel(kernel_size: int, sigma: float) -> jnp.ndarray:
    ax = jnp.linspace(-(kernel_size - 1) / 2., (kernel_size - 1) / 2., kernel_size, dtype=jnp.float64)
    gauss = jnp.exp(-0.5 * jnp.square(ax) / jnp.square(sigma))[..., jnp.newaxis]
    kernel = jnp.outer(gauss, gauss.T)
    return (kernel / jnp.sum(kernel))[..., jnp.newaxis, jnp.newaxis]


def _calculate_ssim(
        img1: jnp.ndarray, img2: jnp.ndarray, kernel: jnp.ndarray, c1: float, c2: float
) -> jnp.ndarray:
    n_channels = img1.shape[-1]

    mu1 = lax.conv_general_dilated(
        img1, kernel, window_strides=(1, 1), padding='VALID', feature_group_count=n_channels,
        dimension_numbers=('NHWC', 'HWIO', 'NHWC'))
    mu2 = lax.conv_general_dilated(
        img2, kernel, window_strides=(1, 1), padding='VALID', feature_group_count=n_channels,
        dimension_numbers=('NHWC', 'HWIO', 'NHWC'))

    mu1_sq = mu1 ** 2
    mu2_sq = mu2 ** 2
    mu12 = mu1 * mu2

    sigma1_sq = lax.conv_general_dilated(
        img1 ** 2, kernel, window_strides=(1, 1),  padding='VALID', feature_group_count=n_channels,
        dimension_numbers=('NHWC', 'HWIO', 'NHWC')
    ) - mu1_sq
    sigma2_sq = lax.conv_general_dilated(
        img2 ** 2, kernel, window_strides=(1, 1), padding='VALID', feature_group_count=n_channels,
        dimension_numbers=('NHWC', 'HWIO', 'NHWC')
    ) - mu2_sq
    sigma12 = lax.conv_general_dilated(
        img1 * img2, kernel, window_strides=(1, 1), padding='VALID', feature_group_count=n_channels,
        dimension_numbers=('NHWC', 'HWIO', 'NHWC')
    ) - mu12

    ssim_map = ((2 * mu12 + c1) * (2 * sigma12 + c2)) / ((mu1_sq + mu2_sq + c1) * (sigma1_sq + sigma2_sq + c2))
    return ssim_map.mean((1, 2, 3))


def ssim(
        img1: jnp.ndarray, img2: jnp.ndarray, crop_border: int, test_y: bool,
        kernel_size: int = 11, sigma: float = 1.5, k1: float = 0.01, k2: float = 0.03,
) -> jnp.ndarray:
    """
    Calculate SSIM between two images.

    Args:
        img1(jnp.ndarray[int, float]): 0 ~ 255 RGB image
        img2(jnp.ndarray[int, float]): 0 ~ 255 RGB image
        crop_border(int): Crop border size.
        test_y(bool): Whether to use Y channel for PSNR calculation.
        kernel_size(int): Gaussian kernel size.
        sigma(float): Gaussian kernel standard deviation.
        k1(float): constant in the SSIM index formula (0.01 in the original paper)
        k2(float): constant in the SSIM index formula (0.03 in the original paper)

    Returns:
        jnp.ndarray: calculated SSIM value
    """
    img1 = preprocess(img1, crop_border=crop_border, to_y=test_y)
    img2 = preprocess(img2, crop_border=crop_border, to_y=test_y)

    kernel = _get_2d_gaussian_kernel(kernel_size=kernel_size, sigma=sigma)
    if img1.shape[-1] != 1:
        kernel = jnp.repeat(kernel, 3, axis=-1)
    c1 = (k1 * 255.) ** 2
    c2 = (k2 * 255.) ** 2

    return _calculate_ssim(img1, img2, kernel, c1, c2)


class SSIM:
    def __init__(
            self, crop_border: int = 0, test_y: bool = False,
            kernel_size: int = 11, sigma: float = 1.5, k1: float = 0.01, k2: float = 0.03
    ):
        """
        Class for calculating SSIM between two images.
        Maybe faster than ssim function when calling multiple times.

        Args:
            crop_border(int): Crop border size.
            test_y(bool): Whether to use Y channel for PSNR calculation.
            kernel_size(int): Gaussian kernel size.
            sigma(float): Gaussian kernel standard deviation.
            k1(float): constant in the SSIM index formula (0.01 in the original paper)
            k2(float): constant in the SSIM index formula (0.03 in the original paper)
        """
        self.kernel_size = kernel_size
        self.sigma = sigma
        kernel = _get_2d_gaussian_kernel(kernel_size=kernel_size, sigma=sigma)
        self.kernel = jnp.repeat(kernel, 3, axis=-1) if not test_y else kernel
        self.k1 = k1
        self.k2 = k2
        self.c1 = (k1 * 255.) ** 2
        self.c2 = (k2 * 255.) ** 2
        self.crop_border = crop_border
        self.test_y = test_y

    def __call__(self, img1: jnp.ndarray, img2: jnp.ndarray) -> jnp.ndarray:
        img1 = preprocess(img1, crop_border=self.crop_border, to_y=self.test_y)
        img2 = preprocess(img2, crop_border=self.crop_border, to_y=self.test_y)
        return _calculate_ssim(img1, img2, self.kernel, self.c1, self.c2)



import logging
import os

import flax.core
import flax.linen as nn
import jax
import jax.lax as lax
import jax.numpy as jnp

from typing import Sequence, Literal
import pickle

import warnings


# TODO: Fix this
logger = logging.getLogger("iqa")


def _load_params(params_in):
    os.makedirs('~/iqa_jax/params', exist_ok=True)
    # os.remove('~/iqa_jax/params/inception_v3.sav')

    if os.path.exists('~/iqa_jax/params/inception_v3.sav'):
        logger.info('Files already exist. Loading from disk.')
        params_bytes = pickle.load(open('~/iqa_jax/params/inception_v3.sav', 'rb'))
        params = flax.core.unfreeze(flax.serialization.from_bytes(params_in, params_bytes))

    else:
        logger.info('Files do not exist. Downloading from tf.keras.')
        warnings.warn('Clearing session. This may cause issues if you are using tf.keras elsewhere.')

        import tensorflow as tf
        tf.keras.backend.clear_session()

        inception = tf.keras.applications.InceptionV3(include_top=False, weights='imagenet')

        params_tf = {}
        for w in inception.weights:
            if w.name.split('/')[0] not in params_tf.keys():
                params_tf[w.name.split('/')[0]] = {}
            params_tf[w.name.split('/')[0]][w.name.split('/')[1]] = jnp.array(w, dtype=jnp.float32)

        del inception
        tf.keras.backend.clear_session()

        # Overwrite the weights with the ones from the pretrained model
        # Params
        for p in params_in['params'].keys():
            if 'conv2d' in p:
                params_in['params'][p]['kernel'] = jnp.asarray(params_tf[p]['kernel:0'], dtype=jnp.float32)
            elif 'batch_normalization' in p:
                params_in['params'][p]['bias'] = jnp.asarray(params_tf[p]['beta:0'], dtype=jnp.float32)

        # Variables
        for p in params_in['batch_stats'].keys():
            params_in['batch_stats'][p]['mean'] = jnp.asarray(params_tf[p]['moving_mean:0'], dtype=jnp.float32)
            params_in['batch_stats'][p]['var'] = jnp.asarray(params_tf[p]['moving_variance:0'], dtype=jnp.float32)

        params_bytes = flax.serialization.to_bytes(params_in)
        pickle.dump(params_bytes, open('~/iqa_jax/params/inception_v3.sav', 'wb'))
        params = flax.core.unfreeze(params_in)

    return params


def load_model():

    def conv_bn_act(
            x, filters: int, kernel_size: Sequence[int], n: int,
            strides: Sequence[int] = (1, 1), padding: str = 'SAME'
    ) -> jnp.ndarray:
        x = nn.Conv(
            features=filters, kernel_size=kernel_size, strides=strides, dtype=jnp.float32,
            padding=padding, use_bias=False, name=f'conv2d_{n}' if n != 0 else 'conv2d')(x)
        x = nn.BatchNorm(
            epsilon=1e-3, use_scale=False, dtype=jnp.float32,
            name=f'batch_normalization_{n}' if n != 0 else 'batch_normalization')(x, use_running_average=True)
        x = nn.relu(x)
        return x

    class InceptionV3(nn.Module):
        @nn.compact
        def __call__(self, x):
            x = x.astype(jnp.float32)
            b, h, w, c = x.shape
            if h != 299 or w != 299:
                x = jax.image.resize(x, (b, 299, 299, c), method='bilinear')

            x = conv_bn_act(x, 32, (3, 3), 0, strides=(2, 2), padding='VALID')
            x = conv_bn_act(x, 32, (3, 3), 1, padding='VALID')
            x = conv_bn_act(x, 64, (3, 3), 2)
            x = nn.max_pool(x, (3, 3), strides=(2, 2), padding='VALID')

            x = conv_bn_act(x, 80, (1, 1), 3, padding='VALID')
            x = conv_bn_act(x, 192, (3, 3), 4, padding='VALID')
            x = nn.max_pool(x, (3, 3), strides=(2, 2), padding='VALID')

            # Mixed 0: 35 x 35 x 256
            branch_1x1 = conv_bn_act(x, 64, (1, 1), 5)

            branch_5x5 = conv_bn_act(x, 48, (1, 1), 6)
            branch_5x5 = conv_bn_act(branch_5x5, 64, (5, 5), 7)

            branch_3x3dbl = conv_bn_act(x, 64, (1, 1), 8)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 9)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 10)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 32, (1, 1), 11)
            x = jnp.concatenate([branch_1x1, branch_5x5, branch_3x3dbl, branch_pool], axis=-1)

            # Mixed 1: 35 x 35 x 288
            branch_1x1 = conv_bn_act(x, 64, (1, 1), 12)

            branch_5x5 = conv_bn_act(x, 48, (1, 1), 13)
            branch_5x5 = conv_bn_act(branch_5x5, 64, (5, 5), 14)

            branch_3x3dbl = conv_bn_act(x, 64, (1, 1), 15)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 16)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 17)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 64, (1, 1), 18)
            x = jnp.concatenate([branch_1x1, branch_5x5, branch_3x3dbl, branch_pool], axis=-1)

            # Mixed 2: 35 x 35 x 288
            branch_1x1 = conv_bn_act(x, 64, (1, 1), 19)

            branch_5x5 = conv_bn_act(x, 48, (1, 1), 20)
            branch_5x5 = conv_bn_act(branch_5x5, 64, (5, 5), 21)

            branch_3x3dbl = conv_bn_act(x, 64, (1, 1), 22)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 23)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 24)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 64, (1, 1), 25)
            x = jnp.concatenate([branch_1x1, branch_5x5, branch_3x3dbl, branch_pool], axis=-1)

            # Mixed 3: 17 x 17 x 768
            branch_3x3 = conv_bn_act(x, 384, (3, 3), 26, strides=(2, 2), padding='VALID')

            branch_3x3dbl = conv_bn_act(x, 64, (1, 1), 27)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 28)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 96, (3, 3), 29, strides=(2, 2), padding='VALID')

            branch_pool = nn.max_pool(x, (3, 3), strides=(2, 2), padding='VALID')
            x = jnp.concatenate([branch_3x3, branch_3x3dbl, branch_pool], axis=-1)

            # Mixed 4: 17 x 17 x 768
            branch_1x1 = conv_bn_act(x, 192, (1, 1), 30)

            branch_7x7 = conv_bn_act(x, 128, (1, 1), 31)
            branch_7x7 = conv_bn_act(branch_7x7, 128, (1, 7), 32)
            branch_7x7 = conv_bn_act(branch_7x7, 192, (7, 1), 33)

            branch_7x7dbl = conv_bn_act(x, 128, (1, 1), 34)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 128, (7, 1), 35)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 128, (1, 7), 36)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 128, (7, 1), 37)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (1, 7), 38)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 39)
            x = jnp.concatenate([branch_1x1, branch_7x7, branch_7x7dbl, branch_pool], axis=-1)

            # Mixed 5: 17 x 17 x 768
            branch_1x1 = conv_bn_act(x, 192, (1, 1), 40)

            branch_7x7 = conv_bn_act(x, 160, (1, 1), 41)
            branch_7x7 = conv_bn_act(branch_7x7, 160, (1, 7), 42)
            branch_7x7 = conv_bn_act(branch_7x7, 192, (7, 1), 43)

            branch_7x7dbl = conv_bn_act(x, 160, (1, 1), 44)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (7, 1), 45)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (1, 7), 46)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (7, 1), 47)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (1, 7), 48)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 49)
            x = jnp.concatenate([branch_1x1, branch_7x7, branch_7x7dbl, branch_pool], axis=-1)

            # Mixed 6: 17 x 17 x 768
            branch_1x1 = conv_bn_act(x, 192, (1, 1), 50)

            branch_7x7 = conv_bn_act(x, 160, (1, 1), 51)
            branch_7x7 = conv_bn_act(branch_7x7, 160, (1, 7), 52)
            branch_7x7 = conv_bn_act(branch_7x7, 192, (7, 1), 53)

            branch_7x7dbl = conv_bn_act(x, 160, (1, 1), 54)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (7, 1), 55)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (1, 7), 56)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 160, (7, 1), 57)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (1, 7), 58)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 59)
            x = jnp.concatenate([branch_1x1, branch_7x7, branch_7x7dbl, branch_pool], axis=-1)

            # Mixed 7: 17 x 17 x 768
            branch_1x1 = conv_bn_act(x, 192, (1, 1), 60)

            branch_7x7 = conv_bn_act(x, 192, (1, 1), 61)
            branch_7x7 = conv_bn_act(branch_7x7, 192, (1, 7), 62)
            branch_7x7 = conv_bn_act(branch_7x7, 192, (7, 1), 63)

            branch_7x7dbl = conv_bn_act(x, 192, (1, 1), 64)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (7, 1), 65)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (1, 7), 66)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (7, 1), 67)
            branch_7x7dbl = conv_bn_act(branch_7x7dbl, 192, (1, 7), 68)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 69)
            x = jnp.concatenate([branch_1x1, branch_7x7, branch_7x7dbl, branch_pool], axis=-1)

            # Mixed 8: 8 x 8 x 1280
            branch_3x3 = conv_bn_act(x, 192, (1, 1), 70)
            branch_3x3 = conv_bn_act(branch_3x3, 320, (3, 3), 71, strides=(2, 2), padding='VALID')

            branch_7x7x3 = conv_bn_act(x, 192, (1, 1), 72)
            branch_7x7x3 = conv_bn_act(branch_7x7x3, 192, (1, 7), 73)
            branch_7x7x3 = conv_bn_act(branch_7x7x3, 192, (7, 1), 74)
            branch_7x7x3 = conv_bn_act(branch_7x7x3, 192, (3, 3), 75, strides=(2, 2), padding='VALID')

            branch_pool = nn.max_pool(x, (3, 3), strides=(2, 2), padding='VALID')
            x = jnp.concatenate([branch_3x3, branch_7x7x3, branch_pool], axis=-1)

            # Mixed 9: 8 x 8 x 2048
            branch_1x1 = conv_bn_act(x, 320, (1, 1), 76)

            branch_3x3 = conv_bn_act(x, 384, (1, 1), 77)
            branch_3x3_1 = conv_bn_act(branch_3x3, 384, (1, 3), 78)
            branch_3x3_2 = conv_bn_act(branch_3x3, 384, (3, 1), 79)
            branch_3x3 = jnp.concatenate([branch_3x3_1, branch_3x3_2], axis=-1)

            branch_3x3dbl = conv_bn_act(x, 448, (1, 1), 80)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 384, (3, 3), 81)
            branch_3x3dbl_1 = conv_bn_act(branch_3x3dbl, 384, (1, 3), 82)
            branch_3x3dbl_2 = conv_bn_act(branch_3x3dbl, 384, (3, 1), 83)
            branch_3x3dbl = jnp.concatenate([branch_3x3dbl_1, branch_3x3dbl_2], axis=-1)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 84)
            x = jnp.concatenate([branch_1x1, branch_3x3, branch_3x3dbl, branch_pool], axis=-1)

            # Mixed 10: 8 x 8 x 2048
            branch_1x1 = conv_bn_act(x, 320, (1, 1), 85)

            branch_3x3 = conv_bn_act(x, 384, (1, 1), 86)
            branch_3x3_1 = conv_bn_act(branch_3x3, 384, (1, 3), 87)
            branch_3x3_2 = conv_bn_act(branch_3x3, 384, (3, 1), 88)
            branch_3x3 = jnp.concatenate([branch_3x3_1, branch_3x3_2], axis=-1)

            branch_3x3dbl = conv_bn_act(x, 448, (1, 1), 89)
            branch_3x3dbl = conv_bn_act(branch_3x3dbl, 384, (3, 3), 90)
            branch_3x3dbl_1 = conv_bn_act(branch_3x3dbl, 384, (1, 3), 91)
            branch_3x3dbl_2 = conv_bn_act(branch_3x3dbl, 384, (3, 1), 92)
            branch_3x3dbl = jnp.concatenate([branch_3x3dbl_1, branch_3x3dbl_2], axis=-1)

            branch_pool = nn.avg_pool(x, (3, 3), strides=(1, 1), padding='SAME')
            branch_pool = conv_bn_act(branch_pool, 192, (1, 1), 93)
            x = jnp.concatenate([branch_1x1, branch_3x3, branch_3x3dbl, branch_pool], axis=-1)

            return x

    module = InceptionV3()
    params_init = flax.core.unfreeze(module.init(jax.random.PRNGKey(0), jnp.ones((1, 299, 299, 3))))
    params = _load_params(params_init)
    return module, params



import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf

from iqa.models import inceptionv3

from absl.testing import absltest
from tqdm import tqdm


jax.config.parse_flags_with_absl()


class InceptionTest(absltest.TestCase):
    def test_weights(self) -> None:
        _, params_flax = inceptionv3.load_model()
        tf.keras.backend.clear_session()
        inception_tf = tf.keras.applications.InceptionV3(include_top=True, weights='imagenet')

        params_tf = {}
        for w in inception_tf.weights:
            if w.name.split('/')[0] not in params_tf.keys():
                params_tf[w.name.split('/')[0]] = {}
            params_tf[w.name.split('/')[0]][w.name.split('/')[1]] = jnp.array(w, dtype=jnp.float32)

        # Params
        for p in tqdm(params_flax['params'].keys(), desc='Params'):
            if 'conv2d' in p:
                p_flax = params_flax['params'][p]['kernel']
                p_tf = params_tf[p]['kernel:0']
            elif 'batch_normalization' in p:
                p_flax = params_flax['params'][p]['bias']
                p_tf = params_tf[p]['beta:0']
            else:
                raise ValueError('Unknown param: {}'.format(p))
            np.testing.assert_allclose(p_flax, p_tf)

        # Variables
        for p in tqdm(params_flax['batch_stats'].keys(), desc='Variables'):
            p_flax = params_flax['batch_stats'][p]['mean']
            p_tf = params_tf[p]['moving_mean:0']
            np.testing.assert_allclose(p_flax, p_tf)

            p_flax = params_flax['batch_stats'][p]['var']
            p_tf = params_tf[p]['moving_variance:0']
            np.testing.assert_allclose(p_flax, p_tf)

    def test_outputs(self) -> None:
        inception_flax, params_flax = inceptionv3.load_model()
        tf.keras.backend.clear_session()
        inception_tf = tf.keras.applications.InceptionV3(include_top=False, weights='imagenet')

        x = np.ones((1, 299, 299, 3), dtype=np.float32)
        y_flax = inception_flax.apply(params_flax, jnp.array(x, jnp.float32))
        y_tf = inception_tf(tf.convert_to_tensor(x, dtype=tf.float32))
        np.testing.assert_allclose(y_flax, y_tf, atol=1e-3)


if __name__ == '__main__':
    absltest.main()
